### Spreadsheet - A Library for reading and writing Spreadsheet Documents.
#
#   Copyright (C) 2008  Hannes Wyss
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#   Contact Information:
#
#     E-Mail: hannes.wyss@gmail.com
#     P-Mail: ywesee GmbH
#             Hannes Wyss
#             Winterthurerstrasse 52
#             8006 Zürich
###           Switzerland

require "parseexcel/parseexcel"
